#ifdef PHONGSHADING
#include <GL/glew.h>
#endif

#include <Math/Math.h>
#include <Geometry/OrthogonalTransformation.h>
#include <GL/gl.h>
#include <GL/GLColorTemplates.h>
#include <GL/GLVertexTemplates.h>
#include <GL/GLContextData.h>
#include <GL/GLGeometryWrappers.h>
#include <GLMotif/Button.h>
#include <GLMotif/Menu.h>
#include <GLMotif/PopupMenu.h>
#include <Vrui/Vrui.h>
#include <Vrui/Application.h>

#include "ply.h"

class VruiDemo:public Vrui::Application,public GLObject
{
  /* Embedded classes: */
private:
  class DataItem:public GLObject::DataItem // Data structure storing OpenGL-dependent application data
  {
    /* Elements: */
  public:
    GLuint textureObjectId; // Texture object ID of some texture
    GLuint displayListId; // Display list ID of some display list
		
#ifdef PHONGSHADING
  // Use to activate/disable shadowShader
  GLuint shadowShaderId;
  GLuint shadowMapUniform;

  // Loading shader function
  GLuint loadShader(char* filename, unsigned int type)
  {
    FILE *pfile;
    GLhandleARB handle;
    const GLcharARB* files[1];
	
    // shader Compilation variable
    GLint result;				// Compilation code result
    GLint errorLoglength ;
    char* errorLogText;
    GLsizei actualErrorLogLength;
	
    char buffer[400000];
    memset(buffer,0,400000);
	
    // This will raise a warning on MS compiler
    pfile = fopen(filename, "rb");
    if(!pfile)
      {
	printf("Sorry, can't open file: '%s'.\n", filename);
	exit(0);
      }
	
    fread(buffer,sizeof(char),400000,pfile);
    //printf("%s\n",buffer);
	
	
    fclose(pfile);
	
    handle = glCreateShader(type);
    if (!handle)
      {
	//We have failed creating the vertex shader object.
	printf("Failed creating vertex shader object from file: %s.",filename);
	exit(0);
      }
	
    files[0] = (const GLcharARB*)buffer;
    glShaderSource(
		   handle, //The handle to our shader
		   1, //The number of files.
		   files, //An array of const char * data, which represents the source code of theshaders
		   NULL);
	
    glCompileShaderARB(handle);
	
    //Compilation checking.
    glGetObjectParameterivARB(handle, GL_OBJECT_COMPILE_STATUS_ARB, &result);
	
    // If an error was detected.
    if (!result)
      {
	//We failed to compile.
	printf("Shader '%s' failed compilation.\n",filename);
		
	//Attempt to get the length of our error log.
	glGetObjectParameterivARB(handle, GL_OBJECT_INFO_LOG_LENGTH_ARB, &errorLoglength);
		
	//Create a buffer to read compilation error message
	errorLogText = (char *)malloc(sizeof(char) * errorLoglength);
		
	//Used to get the final length of the log.
	glGetInfoLogARB(handle, errorLoglength, &actualErrorLogLength, errorLogText);
		
	// Display errors.
	printf("%s\n",errorLogText);
		
	// Free the buffer malloced earlier
	free(errorLogText);
      }
	
    return handle;
  }

  void loadShadowShader()
  {
    GLhandleARB vertexShaderHandle;
    GLhandleARB fragmentShaderHandle;
	
    vertexShaderHandle   = loadShader(
				      "VertexShader-Gooch.c",
				      //"VertexShader-Bump.c",
				      //"VertexShader-Toon.c",
				      //"VertexShader-Phong.c",
				      //"shaders/lambert01vert.glsl",
				      GL_VERTEX_SHADER);
    fragmentShaderHandle = loadShader(
				      "FragmentShader-Gooch.c",
				      //"FragmentShader-Bump.c",
				      //"FragmentShader-Toon.c",
				      //"FragmentShader-Phong.c",
				      //"shaders/woodfrag.glsl",
				      GL_FRAGMENT_SHADER);
	
    shadowShaderId = glCreateProgramObjectARB();
	
    glAttachObjectARB(shadowShaderId,vertexShaderHandle);
    glAttachObjectARB(shadowShaderId,fragmentShaderHandle);
    glLinkProgramARB(shadowShaderId);
  }
#endif

    // ply vertices and faces
    Vertex **vlist;
    Face **flist;
    int num_elems;
    char filename[1024];

    /* Constructors and destructors: */
    DataItem(void)
    {
      /* Create a texture object to hold a texture: */
      glGenTextures(1,&textureObjectId);
			
      /* Create a display list: */
      displayListId=glGenLists(1);

      vlist = 0;
      flist = 0;
      num_elems = 0;
#ifdef PHONGSHADING
      glewInit ();
      loadShadowShader();
#endif
    };
    ~DataItem(void)
    {
      /* Destroy the texture object: */
      glDeleteTextures(1,&textureObjectId);
			
      /* Destroy the display list: */
      glDeleteLists(displayListId,1);
    };

    void computeNormals ();
    void read();

    void render () {
      glColor3f (0.7, 0.7, 0.7);
      for (int j=0; j<num_elems; j++) {
	glBegin (GL_POLYGON);
	for (int k = 0; k < flist[j]->nverts; k++) {
	  glNormal3f (vlist[flist[j]->verts[k]]->nx,
		      vlist[flist[j]->verts[k]]->ny,
		      vlist[flist[j]->verts[k]]->nz);
	  glVertex3f (vlist[flist[j]->verts[k]]->x,
		      vlist[flist[j]->verts[k]]->y,
		      vlist[flist[j]->verts[k]]->z);
	}
	glEnd ();
      }
    }
  };

  /* Elements: */
private:
    char filename[1024];
	
  /* Vrui parameters: */
  GLMotif::PopupMenu* mainMenu; // The program's main menu
	
  /* Private methods: */
  GLMotif::PopupMenu* createMainMenu(void); // Creates the program's main menu

  /* Constructors and destructors: */
public:
  VruiDemo(int& argc,char**& argv,char**& appDefaults); // Initializes the Vrui toolkit and the application
  ~VruiDemo(void); // Shuts down the Vrui toolkit
	
  /* Methods: */
  void setFileName (char* name) {
    strcpy (filename, name);
  }

  virtual void initDisplay(GLContextData& contextData); // Called once upon creation of each OpenGL context
  virtual void initContext(GLContextData& contextData) const;
  virtual void display(GLContextData& contextData) const; // Called for every eye and every window on every frame
  virtual void frame(void); // Called exactly once per frame
  void resetNavigationCallback(Misc::CallbackData* cbData); // Method to reset the Vrui navigation transformation to its default
};

/*************************
Methods of class VruiDemo:
*************************/

void VruiDemo::DataItem::computeNormals () {
  int j, k;
  double normal[3], v1[3], v2[3], length;

  if (fabs (vlist[flist[0]->verts[0]]->nx +
	    vlist[flist[0]->verts[0]]->ny +
	    vlist[flist[0]->verts[0]]->nz) > 0.05) {
#ifdef DEBUG
    printf ("normal information present; do not need to calculate\n");
#endif
    return;
  }

  for (j=0; j<num_elems; j++) {
    // compute normal for face first
    v1[0] = vlist[flist[j]->verts[0]]->x - vlist[flist[j]->verts[1]]->x;
    v1[1] = vlist[flist[j]->verts[0]]->y - vlist[flist[j]->verts[1]]->y;
    v1[2] = vlist[flist[j]->verts[0]]->z - vlist[flist[j]->verts[1]]->z;
    v2[0] = vlist[flist[j]->verts[0]]->x - vlist[flist[j]->verts[2]]->x;
    v2[1] = vlist[flist[j]->verts[0]]->y - vlist[flist[j]->verts[2]]->y;
    v2[2] = vlist[flist[j]->verts[0]]->z - vlist[flist[j]->verts[2]]->z;
    normal[0] = v1[1] * v2[2] - v1[2] * v2[1];
    normal[1] = v1[2] * v2[0] - v1[0] * v2[2];
    normal[2] = v1[0] * v2[1] - v1[1] * v2[0];

#ifdef DEBUG
    printf ("V1 vector: (%lf,%lf,%lf)\n",
	    v1[0], v1[1], v1[2]);
    printf ("V2 vector: (%lf,%lf,%lf)\n",
	    v2[0], v2[1], v2[2]);
    printf ("Normal vector: (%lf,%lf,%lf)\n",
	    normal[0], normal[1], normal[2]);
#endif
  
    // update normal vector for vertices
    for (k = 0; k < flist[j]->nverts; k++) {
      vlist[flist[j]->verts[k]]->nx += normal[0];
      vlist[flist[j]->verts[k]]->ny += normal[1];
      vlist[flist[j]->verts[k]]->nz += normal[2];
    }
  }

  // normalize normals
  for (j=0; j<num_elems; j++) {
    for (k = 0; k < flist[j]->nverts; k++) {
      length = sqrt (vlist[flist[j]->verts[k]]->nx *
		     vlist[flist[j]->verts[k]]->nx +
		     vlist[flist[j]->verts[k]]->ny *
		     vlist[flist[j]->verts[k]]->ny +
		     vlist[flist[j]->verts[k]]->nz *
		     vlist[flist[j]->verts[k]]->nz);
      vlist[flist[j]->verts[k]]->nx /= length;
      vlist[flist[j]->verts[k]]->ny /= length;
      vlist[flist[j]->verts[k]]->nz /= length;
    }
  }
}

/******************************************************************************
Read in a PLY file.
******************************************************************************/

void VruiDemo::DataItem::read()
{
  int i,j,k;
  PlyFile *ply;
  int nelems;
  char **elist;
  int file_type;
  float version;
  int nprops;
  PlyProperty **plist;
  char *elem_name;
  int num_comments;
  char **comments;
  int num_obj_info;
  char **obj_info;

  /* open a PLY file for reading */
  ply = ply_open_for_reading((char*)filename, &nelems, &elist, &file_type, &version);

#ifdef DEBUG
  /* print what we found out about the file */
  printf ("version %f\n", version);
  printf ("type %d\n", file_type);
#endif

  /* go through each kind of element that we learned is in the file */
  /* and read them */

  for (i = 0; i < nelems; i++) {

    /* get the description of the first element */
    elem_name = elist[i];
    plist = ply_get_element_description (ply, elem_name, &num_elems, &nprops);

#ifdef DEBUG
    /* print the name of the element, for debugging */
    printf ("element %s %d\n", elem_name, num_elems);
#endif

    /* if we're on vertex elements, read them in */
    if (equal_strings ("vertex", elem_name)) {

      /* create a vertex list to hold all the vertices */
      vlist = (Vertex **) malloc (sizeof (Vertex *) * num_elems);

      /* set up for getting vertex elements */

      ply_get_property (ply, elem_name, &vert_props[0]);
      ply_get_property (ply, elem_name, &vert_props[1]);
      ply_get_property (ply, elem_name, &vert_props[2]);
      ply_get_property (ply, elem_name, &vert_props[3]);
      ply_get_property (ply, elem_name, &vert_props[4]);
      ply_get_property (ply, elem_name, &vert_props[5]);

      /* grab all the vertex elements */
      for (j = 0; j < num_elems; j++) {

	/* grab and element from the file */
	vlist[j] = (Vertex *) malloc (sizeof (Vertex));
	ply_get_element (ply, (void *) vlist[j]);

#ifdef DEBUG
	/* print out vertex x,y,z for debugging */
	printf ("vertex: %g %g %g\n", vlist[j]->x, vlist[j]->y, vlist[j]->z);
#endif
      }
    }

    /* if we're on face elements, read them in */
    if (equal_strings ("face", elem_name)) {

      /* create a list to hold all the face elements */
      flist = (Face **) malloc (sizeof (Face *) * num_elems);

      /* set up for getting face elements */

      ply_get_property (ply, elem_name, &face_props[0]);
      ply_get_property (ply, elem_name, &face_props[1]);

      /* grab all the face elements */
      for (j = 0; j < num_elems; j++) {

	/* grab and element from the file */
	flist[j] = (Face *) malloc (sizeof (Face));
	ply_get_element (ply, (void *) flist[j]);

#ifdef DEBUG
	/* print out face info, for debugging */
	printf ("face: %d, list = ", flist[j]->intensity);
	for (k = 0; k < flist[j]->nverts; k++)
	  printf ("%d ", flist[j]->verts[k]);
	printf ("\n");
#endif
      }
    }

#ifdef DEBUG
    /* print out the properties we got, for debugging */
    for (j = 0; j < nprops; j++)
      printf ("property %s\n", plist[j]->name);
#endif
  }

  /* grab and print out the comments in the file */
  comments = ply_get_comments (ply, &num_comments);
#ifdef DEBUG
  for (i = 0; i < num_comments; i++)
    printf ("comment = '%s'\n", comments[i]);
#endif

  /* grab and print out the object information */
  obj_info = ply_get_obj_info (ply, &num_obj_info);
#ifdef DEBUG
  for (i = 0; i < num_obj_info; i++)
    printf ("obj_info = '%s'\n", obj_info[i]);
#endif

  /* close the PLY file */
  ply_close (ply);

  computeNormals ();
}

GLMotif::PopupMenu* VruiDemo::createMainMenu(void)
{
  /* Create a popup shell to hold the main menu: */
  GLMotif::PopupMenu* mainMenuPopup=new GLMotif::PopupMenu("MainMenuPopup",Vrui::getWidgetManager());
  mainMenuPopup->setBorderWidth(0.0f);
  mainMenuPopup->setBorderType(GLMotif::Widget::RAISED);
  mainMenuPopup->setBorderColor(Vrui::getUiBgColor());
  mainMenuPopup->setBackgroundColor(Vrui::getUiBgColor());
  mainMenuPopup->setForegroundColor(Vrui::getUiFgColor());
  mainMenuPopup->setMarginWidth(Vrui::getUiSize());
  mainMenuPopup->setTitleSpacing(Vrui::getUiSize());
  mainMenuPopup->setTitle("Vrui Demonstration",Vrui::getUiFont());
	
  /* Create the main menu itself: */
  GLMotif::Menu* mainMenu=new GLMotif::Menu("MainMenu",mainMenuPopup,false);
  mainMenu->setBorderWidth(0.0f);
  mainMenu->setOrientation(GLMotif::Menu::VERTICAL);
  mainMenu->setNumMinorWidgets(1);
  mainMenu->setMarginWidth(0.0f);
  mainMenu->setSpacing(Vrui::getUiSize());
	
  /* Create a button: */
  GLMotif::Button* resetNavigationButton=new GLMotif::Button("ResetNavigationButton",mainMenu,"Reset Navigation",Vrui::getUiFont());
	
  /* Add a callback to the button: */
  resetNavigationButton->getSelectCallbacks().add(this,&VruiDemo::resetNavigationCallback);
	
  /* Finish building the main menu: */
  mainMenu->manageChild();
	
  return mainMenuPopup;
}

VruiDemo::VruiDemo(int& argc,char**& argv,char**& appDefaults)
  :Vrui::Application(argc,argv,appDefaults),
   mainMenu(0)
{
  /* Create the user interface: */
  mainMenu=createMainMenu();
	
  /* Install the main menu: */
  Vrui::setMainMenu(mainMenu);
	
  /* Set the navigation transformation: */
  resetNavigationCallback(0);
}

VruiDemo::~VruiDemo(void)
{
  delete mainMenu;
}

void VruiDemo::initDisplay(GLContextData& contextData)
{
  /*********************************************************************
	This method is called once for every OpenGL context created for the
	program. It must not change application state, but only create
	OpenGL-dependent application data and store them in the GLContextData
	object for retrieval in the display method.
  *********************************************************************/

}
	
void VruiDemo::initContext(GLContextData& contextData) const
{
  /* Create context data item and store it in the GLContextData object: */
  DataItem* dataItem=new DataItem;
  int left, right, bottom, top, near, far;
  contextData.addDataItem(this,dataItem);
  strcpy (dataItem->filename, filename);
  dataItem->read ();

  /* Now is the time to upload all needed texture data: */
  glBindTexture(GL_TEXTURE_2D,dataItem->textureObjectId);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_REPEAT);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_REPEAT);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
  GLfloat texImage[4][4]=
    {
      {0.0f, 0.0f, 1.0f, 1.0f},
      {0.0f, 0.0f, 1.0f, 1.0f},
      {1.0f, 1.0f, 0.0f, 0.0f},
      {1.0f, 1.0f, 0.0f, 0.0f}
    };
  glTexImage2D(GL_TEXTURE_2D,0,GL_RGBA,4,4,0,GL_LUMINANCE,GL_FLOAT,texImage);
  glBindTexture(GL_TEXTURE_2D,0);
	
  /* Now is also the time to upload all display lists' contents: */
  glNewList(dataItem->displayListId,GL_COMPILE);

  //glScale (2.5, 2.5, 2.5);
  glScale (9.5, 9.5, 9.5);
  //glRotatef (10.0, 1.0, 0.0, 0.0);

  /* Draw the scene */
  //glPushAttrib(GL_LIGHTING_BIT);

  // draw the entire geometry by looping through the faces which in
  // turn reference the vertices
  dataItem->render ();
  //glPopAttrib();
	
  /* Finish the display list: */
  glEndList();
}

void VruiDemo::display(GLContextData& contextData) const
{
  /*********************************************************************
	This method is called once for every eye in every window on every
	frame. It must not change application state, as it is called an
	unspecified number of times. It also must not clear the screen or
	initialize the OpenGL transformation matrices. When this method is
	called, Vrui will already have rendered its own state (menus etc.) and
	have set up the transformation matrices so that all rendering in this
	method happens in navigation (i.e., model) coordinates.
  *********************************************************************/
	
  /* Get the OpenGL-dependent application data from the GLContextData object: */
  DataItem* dataItem=contextData.retrieveDataItem<DataItem>(this);
	
  /* Insert generic OpenGL code here... */
  // ...
	
  /* Use the texture uploaded in the initDisplay() method: */
  glBindTexture(GL_TEXTURE_2D,dataItem->textureObjectId);
  // ...
  glBindTexture(GL_TEXTURE_2D,0);
	
  /* Set up the animation transformation: */
  glPushMatrix();
	
  GLfloat mat_ambient[] = { 0.7, 0.7, 0.7, 1.0 };
  GLfloat mat_diffuse[] = { 0.8, 0.8, 0.8, 1.0 };
  GLfloat mat_specular[] = { 0.5, 0.5, 0.5, 0.5 };
  GLfloat high_shininess[] = { 50.0 };

  glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, mat_ambient);
  glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, mat_diffuse);
  glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, mat_specular);
  glMaterialfv(GL_FRONT_AND_BACK, GL_SHININESS, high_shininess);

  glDepthFunc(GL_LESS);
  glEnable(GL_DEPTH_TEST);
  glEnable(GL_COLOR_MATERIAL);
  glEnable (GL_BLEND);
  glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
  glLightModeli (GL_LIGHT_MODEL_TWO_SIDE, GL_TRUE);

#ifdef PHONGSHADING
  // Use the shadow shader
  glUseProgram(dataItem->shadowShaderId);
#endif

  glDisable (GL_CULL_FACE);
  /* Call the display list created in the initDisplay() method: */
  glCallList(dataItem->displayListId);
  glEnable (GL_CULL_FACE);
	
#ifdef PHONGSHADING
  // Disable shadow shader
  glUseProgram(0);
#endif

  /* Reset the animation transformation: */
  glPopMatrix();
}

void VruiDemo::frame(void)
{
  /*********************************************************************
	This function is called exactly once per frame, no matter how many
	eyes or windows exist. It is the perfect place to change application
	state (run simulations, animate models, etc.).
  *********************************************************************/
	
  /* Get the time since the last frame: */
  double frameTime=Vrui::getCurrentFrameTime();
	
}

void VruiDemo::resetNavigationCallback(Misc::CallbackData* cbData)
{
  /* Reset the Vrui navigation transformation: */
  Vrui::NavTransform t=Vrui::NavTransform::identity;
  t*=Vrui::NavTransform::translateFromOriginTo(Vrui::getDisplayCenter());
  t*=Vrui::NavTransform::scale(Vrui::getInchFactor());
  Vrui::setNavigationTransformation(t);
	
  /*********************************************************************
	Now the coordinate system's origin is in the middle of the
	environment, e.g., in the middle of the workbench, and one coordinate
	unit is one inch long.
  *********************************************************************/
}

int main(int argc,char* argv[])
{
  /* Create an application object: */
  char** appDefaults=0; // This is an additional parameter no one ever uses
  VruiDemo app(argc,argv,appDefaults);
	
  // read the ply file
  if (argc == 2)
    app.setFileName (argv[1]);
  else
    app.setFileName ("icosahedron.ply");

  /* Run the Vrui main loop: */
  app.run();
	
  /* Exit to OS: */
  return 0;
}
